package gov.usgs.earthquake.nshmp.netcdf.reader;

import java.io.IOException;
import java.util.List;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.netcdf.Netcdf;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfGroundMotions;
import gov.usgs.earthquake.nshmp.netcdf.data.BoundingData;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfShape.IndexKey;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;
import gov.usgs.earthquake.nshmp.netcdf.reader.NetcdfUtils.Key;

import ucar.ma2.Array;
import ucar.ma2.DataType;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.dataset.NetcdfDatasets;

/**
 * Creates the bounding locations and data associated with a specific site for
 * ground motions.
 *
 * @author U.S. Geological Survey
 */
public class BoundingReaderGroundMotions extends BoundingReader<XySequence> {

  public static final double PGA_VALUE = 0.001;
  public static final double PGV_VALUE = 0.0001;

  public BoundingReaderGroundMotions(NetcdfGroundMotions netcdf, Location site) {
    super(netcdf, site);
  }

  @Override
  StaticData<XySequence> calculateTargetData(
      List<BoundingLocation> boundingLocations,
      BoundingData<XySequence> boundingData,
      double fracLon,
      double fracLat) {
    var westTarget = getTargetData(
        boundingData.get(boundingLocations.get(0).location),
        boundingData.get(boundingLocations.get(1).location),
        fracLat);

    var eastTarget = getTargetData(
        boundingData.get(boundingLocations.get(3).location),
        boundingData.get(boundingLocations.get(2).location),
        fracLat);

    return getTargetData(westTarget, eastTarget, fracLon);
  }

  @Override
  BoundingData<XySequence> extractDataAt(
      Netcdf<XySequence> netcdf,
      List<BoundingLocation> boundingLocations,
      int idxLonLL,
      int idxLatLL) {
    try (var ncd = NetcdfDatasets.openDataset(netcdf.netcdfPath().toString())) {
      var netcdfData = netcdf.netcdfData();
      var boundingData = BoundingData.<XySequence> builder();
      var targetGroup = ncd.getRootGroup();
      var netcdfShape = netcdf.netcdfShape();

      // Build origin array, e.g. [lon, lat, 0, 0]
      var targetOrigin = netcdfShape.buildShape()
          .add(IndexKey.LATITUDE, idxLatLL)
          .add(IndexKey.LONGITUDE, idxLonLL)
          .build();

      // Build target shape array, e.g. [2, 2, nImt, nSiteClass]
      var targetShape = netcdfShape.buildShape()
          .add(IndexKey.SITE_CLASS, netcdfData.siteClasses().size())
          .add(IndexKey.LATITUDE, 2)
          .add(IndexKey.LONGITUDE, 2)
          .add(IndexKey.IMT, netcdfData.imts().size())
          .build();

      var groundMotionArray = targetGroup
          .findVariableLocal(Key.GROUND_MOTION)
          .read(targetOrigin, targetShape);

      // Main data shape, e.g. [1, 1, nImt, nSiteClass]
      var shape = netcdfShape.buildShape()
          .add(IndexKey.SITE_CLASS, netcdfData.siteClasses().size())
          .add(IndexKey.LATITUDE, 1)
          .add(IndexKey.LONGITUDE, 1)
          .add(IndexKey.IMT, netcdfData.imts().size())
          .build();

      for (var boundingLocation : boundingLocations) {
        boundingData.put(
            boundingLocation.location,
            mapDataFromArray(netcdf, groundMotionArray.section(boundingLocation.origin, shape)));
      }

      return boundingData.build();
    } catch (IOException | InvalidRangeException e) {
      throw new RuntimeException(
          "Could not read Netcdf file [" + netcdf.netcdfPath() + "]. " + e.getMessage());
    }
  }

  @Override
  StaticData<XySequence> getTargetData(
      StaticData<XySequence> d1,
      StaticData<XySequence> d2,
      double frac) {
    NetcdfUtils.checkBoundingGroundMotion(d1, d2);
    return frac == 0.0 ? d1
        : frac == 1.0 ? d2 : NetcdfUtils.linearInterpolateGroundMotions(d1, d2, frac);
  }

  @Override
  StaticData<XySequence> mapDataFromArray(
      Netcdf<XySequence> netcdf,
      Array array) {
    var netcdfData = netcdf.netcdfData();
    var staticData = StaticData.<XySequence> builder();

    for (int iSiteClass = 0; iSiteClass < netcdfData.siteClasses().size(); iSiteClass++) {
      var siteClass = netcdfData.siteClasses().get(iSiteClass);

      var imts = netcdfData.imts();
      var periods = imts.stream()
          .mapToDouble(imt -> {
            if (imt == Imt.PGA) {
              return PGA_VALUE;
            } else if (imt == Imt.PGV) {
              return PGV_VALUE;
            }
            return imt.period();
          })
          .toArray();

      // Build origin array, e.g [0, siteClass]
      var origin = netcdf.netcdfShape().buildShape()
          .add(IndexKey.IMT, 0)
          .add(IndexKey.SITE_CLASS, iSiteClass)
          .reduce()
          .build();

      // Build shape array, e.g. [nImts, 1]
      var shape = netcdf.netcdfShape().buildShape()
          .add(IndexKey.IMT, imts.size())
          .add(IndexKey.SITE_CLASS, 1)
          .reduce()
          .build();

      try {
        var xySequence = XySequence.create(
            periods,
            (double[]) array.section(origin, shape).reduce().get1DJavaArray(DataType.DOUBLE));

        staticData.put(siteClass, xySequence);
      } catch (InvalidRangeException e) {
        throw new RuntimeException(e.getMessage());
      }
    }

    return staticData.build();
  }

  @Override
  BoundingData<XySequence> setBoundingData(
      Netcdf<XySequence> netcdf,
      Location site,
      List<BoundingLocation> boundingLocations) {
    var netcdfGroundMotions = (NetcdfGroundMotions) netcdf;
    var netcdfData = netcdfGroundMotions.netcdfData();
    var longitudes = netcdfData.longitudes();
    var latitudes = netcdfData.latitudes();
    var idxLonLL = NetcdfUtils.getIdxLTEQ(longitudes, site.longitude);
    var idxLatLL = NetcdfUtils.getIdxLTEQ(latitudes, site.latitude);

    var groundMotions =
        extractDataAt(netcdfGroundMotions, boundingLocations, idxLonLL, idxLatLL);
    var fracLon = NetcdfUtils.calcGridFrac(longitudes, idxLonLL, site.longitude);
    var fracLat = NetcdfUtils.calcGridFrac(latitudes, idxLatLL, site.latitude);

    var builder = BoundingData.<XySequence> builder();
    groundMotions.forEach((key, value) -> builder.put(key, value));
    builder.put(
        site,
        calculateTargetData(boundingLocations, groundMotions, fracLon, fracLat))
        .build();
    var boundingData = builder.build();

    NetcdfUtils.checkBoundingGroundMotions(boundingData, boundingLocations.get(0).location);
    return boundingData;
  }
}
