package gov.usgs.earthquake.nshmp.netcdf.reader;

import java.io.IOException;
import java.util.List;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.netcdf.Netcdf;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.data.BoundingData;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfShape.IndexKey;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticDataHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.reader.NetcdfUtils.Key;

import ucar.ma2.Array;
import ucar.ma2.DataType;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.dataset.NetcdfDatasets;

/**
 * Creates the bounding locations and data associated with a specific site for
 * hazard curves.
 *
 * @author U.S. Geological Survey
 */
public class BoundingReaderHazardCurves extends BoundingReader<StaticDataHazardCurves> {

  public BoundingReaderHazardCurves(NetcdfHazardCurves netcdf, Location site) {
    super(netcdf, site);
  }

  @Override
  StaticData<StaticDataHazardCurves> calculateTargetData(
      List<BoundingLocation> boundingLocations,
      BoundingData<StaticDataHazardCurves> hazards,
      double fracLon,
      double fracLat) {
    var westTarget = getTargetData(
        hazards.get(boundingLocations.get(0).location),
        hazards.get(boundingLocations.get(1).location),
        fracLat);

    var eastTarget = getTargetData(
        hazards.get(boundingLocations.get(3).location),
        hazards.get(boundingLocations.get(2).location),
        fracLat);

    return getTargetData(westTarget, eastTarget, fracLon);
  }

  @Override
  BoundingData<StaticDataHazardCurves> extractDataAt(
      Netcdf<StaticDataHazardCurves> netcdf,
      List<BoundingLocation> boundingLocations,
      int idxLonLL,
      int idxLatLL) {
    try (var ncd = NetcdfDatasets.openDataset(netcdf.netcdfPath().toString())) {
      var netcdfData = ((NetcdfHazardCurves) netcdf).netcdfData();
      var boundingData = BoundingData.<StaticDataHazardCurves> builder();
      var targetGroup = ncd.getRootGroup();
      var netcdfShape = netcdf.netcdfShape();

      // Build origin array, e.g. [lon, lat, 0, 0, 0]
      var targetOrigin = netcdfShape.buildShape()
          .add(IndexKey.LATITUDE, idxLatLL)
          .add(IndexKey.LONGITUDE, idxLonLL)
          .build();

      // Build target shape array, e.g. [2, 2, nIML, nIMT, nSiteClass]
      var targetShape = netcdfShape.buildShape()
          .add(IndexKey.LONGITUDE, 2)
          .add(IndexKey.LATITUDE, 2)
          .add(IndexKey.IML, netcdfData.nIml())
          .add(IndexKey.IMT, netcdfData.imts().size())
          .add(IndexKey.SITE_CLASS, netcdfData.siteClasses().size())
          .build();

      var aHazards = targetGroup.findVariableLocal(Key.HAZARD).read(targetOrigin, targetShape);

      // Build final shape array, e.g. [1, 1, nIML, nIMT, nSiteClass]
      var shape = netcdfShape.buildShape()
          .add(IndexKey.LONGITUDE, 1)
          .add(IndexKey.LATITUDE, 1)
          .add(IndexKey.IML, netcdfData.nIml())
          .add(IndexKey.IMT, netcdfData.imts().size())
          .add(IndexKey.SITE_CLASS, netcdfData.siteClasses().size())
          .build();

      for (var boundingLocation : boundingLocations) {
        boundingData.put(
            boundingLocation.location,
            mapDataFromArray(netcdf, aHazards.section(boundingLocation.origin, shape)));
      }

      return boundingData.build();
    } catch (IOException | InvalidRangeException e) {
      throw new RuntimeException("Could not read Netcdf file [" + netcdf.netcdfPath() + "]");
    }
  }

  @Override
  StaticData<StaticDataHazardCurves> getTargetData(
      StaticData<StaticDataHazardCurves> d1,
      StaticData<StaticDataHazardCurves> d2,
      double frac) {
    NetcdfUtils.checkBoundingHazard(d1, d2);
    return frac == 0.0 ? d1
        : frac == 1.0 ? d2 : NetcdfUtils.linearInterpolateHazardCurves(d1, d2, frac);
  }

  @Override
  StaticData<StaticDataHazardCurves> mapDataFromArray(
      Netcdf<StaticDataHazardCurves> netcdf,
      Array hazards) {
    var netcdfData = ((NetcdfHazardCurves) netcdf).netcdfData();
    var vsImtHazardMap = StaticData.<StaticDataHazardCurves> builder();

    for (int iSiteClass = 0; iSiteClass < netcdfData.siteClasses().size(); iSiteClass++) {
      var siteClass = netcdfData.siteClasses().get(iSiteClass);
      var imtHazardMap = StaticDataHazardCurves.builder();

      for (int iImt = 0; iImt < netcdfData.imts().size(); iImt++) {
        var imt = netcdfData.imts().get(iImt);

        // Build origin array, e.g [0, imt, siteClass]
        var origin = netcdf.netcdfShape().buildShape()
            .add(IndexKey.IML, 0)
            .add(IndexKey.IMT, iImt)
            .add(IndexKey.SITE_CLASS, iSiteClass)
            .reduce()
            .build();

        // Build shape array, e.g. [nIML, 1, 1]
        var shape = netcdf.netcdfShape().buildShape()
            .add(IndexKey.IML, netcdfData.nIml())
            .add(IndexKey.IMT, 1)
            .add(IndexKey.SITE_CLASS, 1)
            .reduce()
            .build();

        try {
          var xySequence = XySequence.create(
              netcdfData.imls().get(imt),
              (double[]) hazards.section(origin, shape).reduce().get1DJavaArray(DataType.DOUBLE));

          imtHazardMap.put(imt, xySequence);
        } catch (InvalidRangeException e) {
          throw new RuntimeException(e.getMessage());
        }
      }

      vsImtHazardMap.put(siteClass, imtHazardMap.build());
    }

    return vsImtHazardMap.build();
  }

  @Override
  BoundingData<StaticDataHazardCurves> setBoundingData(
      Netcdf<StaticDataHazardCurves> netcdf,
      Location site,
      List<BoundingLocation> boundingLocations) {
    var netcdfHazard = (NetcdfHazardCurves) netcdf;
    var netcdfData = netcdfHazard.netcdfData();
    var longitudes = netcdfData.longitudes();
    var latitudes = netcdfData.latitudes();
    var idxLonLL = NetcdfUtils.getIdxLTEQ(longitudes, site.longitude);
    var idxLatLL = NetcdfUtils.getIdxLTEQ(latitudes, site.latitude);

    var hazards = extractDataAt(netcdfHazard, boundingLocations, idxLonLL, idxLatLL);
    var fracLon = NetcdfUtils.calcGridFrac(longitudes, idxLonLL, site.longitude);
    var fracLat = NetcdfUtils.calcGridFrac(latitudes, idxLatLL, site.latitude);

    var builder = BoundingData.<StaticDataHazardCurves> builder();
    hazards.forEach((key, value) -> builder.put(key, value));
    builder.put(
        site,
        calculateTargetData(boundingLocations, hazards, fracLon, fracLat))
        .build();
    var boundingData = builder.build();

    NetcdfUtils.checkBoundingHazards(boundingData, boundingLocations.get(0).location);
    return boundingData;
  }
}
