package gov.usgs.earthquake.nshmp.netcdf.reader;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.google.common.reflect.TypeToken;
import com.google.gson.Gson;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfData;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfShape;
import gov.usgs.earthquake.nshmp.netcdf.data.ScienceBaseMetadata;
import gov.usgs.earthquake.nshmp.netcdf.reader.NetcdfUtils.Key;

import ucar.nc2.Group;

/**
 * Read in NetCDF file.
 *
 * @author U.S. Geological Survey
 */
public abstract class Reader {

  Group targetGroup;
  NetcdfData netcdfData;
  private NetcdfShape netcdfShape;

  static final String FLAG_KEY = "flag_key";
  static final Gson GSON = new Gson();

  public Reader(Group targetGroup) throws IOException {
    this.targetGroup = targetGroup;
    netcdfData = readData(targetGroup);
    netcdfShape = buildNetcdfShape(targetGroup);
  }

  public NetcdfShape netcdfShape() {
    return netcdfShape;
  }

  public NetcdfData netcdfData() {
    return netcdfData;
  }

  abstract NetcdfShape buildNetcdfShape(Group group);

  NetcdfData readData(Group targetGroup) throws IOException {
    var vSiteClass = targetGroup.findVariableLocal(Key.SITE_CLASS);
    var vs30s = NetcdfUtils.getDoubleArray(targetGroup, Key.VS30);
    var vImts = targetGroup.findVariableLocal(Key.IMT);

    var mapTypeToken = new TypeToken<Map<String, String>>() {}.getType();

    // get list of IMT enums
    List<Imt> imtList = new ArrayList<Imt>();
    Map<String, String> imtFlagKeys = GSON
        .fromJson(vImts.findAttribute(FLAG_KEY)
            .getValues().toString(), mapTypeToken);

    var imtArray = vImts.read();
    for (int i = 0; i < vImts.getSize(); i++) {
      var imtInt = imtArray.getInt(i);
      var imtString = imtFlagKeys.get(Integer.toString(imtInt)).toString();
      imtList.add(Imt.valueOf(imtString));
    }
    var imts = List.copyOf(imtList);

    // get list of SiteClass enums and build vs30 map
    var scList = new ArrayList<NehrpSiteClass>();
    var vsMap = new HashMap<NehrpSiteClass, Double>();
    var siteClassArray = vSiteClass.read();
    Map<String, String> siteClassFlagKeys = GSON
        .fromJson(vSiteClass.findAttribute(FLAG_KEY)
            .getValues().toString(), mapTypeToken);

    for (int i = 0; i < vSiteClass.getSize(); i++) {
      var scInt = siteClassArray.getInt(i);
      var scString = siteClassFlagKeys.get(Integer.toString(scInt)).toString();
      var siteClass = NehrpSiteClass.valueOf(scString);
      scList.add(siteClass);
      vsMap.put(siteClass, vs30s[i]);
    }
    var siteClasses = List.copyOf(scList);
    var vs30Map = Map.copyOf(vsMap);

    var latitudes = Arrays.stream(NetcdfUtils.getDoubleArray(targetGroup, Key.LAT))
        // TODO: Dynamic set location precision from NetCDF
        .map(lat -> Maths.round(lat, 3))
        .toArray();

    var longitudes = Arrays.stream(NetcdfUtils.getDoubleArray(targetGroup, Key.LON))
        // TODO: Dynamic set location precision from NetCDF
        .map(lon -> Maths.round(lon, 3))
        .toArray();

    return NetcdfData.builder()
        .imts(imts)
        .latitudes(latitudes)
        .longitudes(longitudes)
        .scienceBaseMetadata(new ScienceBaseMetadata(targetGroup))
        .siteClasses(siteClasses)
        .vs30Map(vs30Map)
        .build();
  }
}
