package gov.usgs.earthquake.nshmp.netcdf.www;

import java.nio.file.Path;

import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfDataType;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfGroundMotions;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfHazardCurves;
import gov.usgs.earthquake.nshmp.www.NshmpMicronautServlet;

import io.micronaut.context.annotation.Value;
import io.micronaut.context.event.StartupEvent;
import io.micronaut.core.annotation.Nullable;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.PathVariable;
import io.micronaut.http.annotation.QueryValue;
import io.micronaut.runtime.event.annotation.EventListener;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.inject.Inject;

/**
 * Micronaut controller for getting static hazards or ground motions from a
 * NetCDF file.
 *
 * @see NetcdfService
 *
 * @author U.S. Geological Survey
 */
@Tag(name = "Static Data")
@Controller("/curves")
public class NetcdfController {

  @Inject
  private NshmpMicronautServlet servlet;

  @Value("${nshmp-ws-static.netcdf-file}")
  Path netcdfPath;

  NetcdfService service;

  /**
   * Read in data type and return the appropriate service to use.
   */
  @EventListener
  void startup(StartupEvent event) {
    var dataType = NetcdfDataType.getDataType(netcdfPath);

    switch (dataType) {
      case GROUND_MOTIONS:
        var netcdfGroundMotions = new NetcdfGroundMotions(netcdfPath);
        service = new NetcdfServiceGroundMotions(netcdfGroundMotions);
        break;
      case HAZARD_CURVES:
        var netcdfHazard = new NetcdfHazardCurves(netcdfPath);
        service = new NetcdfServiceHazardCurves(netcdfHazard);
        break;
      default:
        throw new RuntimeException("Data type [" + dataType + "] not supported");
    }
  }

  /**
   * GET method to return a static curve using URL query.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class (optional)
   */
  @Operation(
      summary = "Returns curve(s) given a longitude, latitude, and/or a site class",
      description = "Retrieve static curve(s) from a NSHM NetCDF file.\n\n" +
          "For supported longitudes, latitudes, and site classes see the usage information.",
      operationId = "netcdf_data_doGetHazard")
  @ApiResponse(
      description = "Returns a static curve from the NetCDF file",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "{?longitude,latitude,siteClass}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGet(
      HttpRequest<?> request,
      @Schema(required = true) @QueryValue @Nullable Double longitude,
      @Schema(required = true) @QueryValue @Nullable Double latitude,
      @QueryValue @Nullable NehrpSiteClass siteClass) {
    var query = new Query(longitude, latitude, siteClass);
    return service.handleServiceCall(request, query);
  }

  /**
   * GET method to return static curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class
   */
  @Operation(
      summary = "Returns static curves given a longitude, latitude, and site class.",
      description = "Retrieve static curves from a NetCDF file.\n\n" +
          "For supported longitudes, latitudes, and site classes see the usage information.",
      operationId = "netcdf_data_doGetHazardSlashWithSiteClass")
  @ApiResponse(
      description = "Returns static curves from the NetCDF file",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/{longitude}/{latitude}/{siteClass}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetSlashBySite(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable @Nullable Double longitude,
      @Schema(required = true) @PathVariable @Nullable Double latitude,
      @Schema(required = true) @PathVariable @Nullable NehrpSiteClass siteClass) {
    return doGet(request, longitude, latitude, siteClass);
  }

  /**
   * GET method to return hazard curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   */
  @Operation(
      summary = "Returns static curves given a longitude and latitude.",
      description = "Retrieve static curves from a NetCDF file.\n\n" +
          "For supported longitudes and latitudes see the usage information.",
      operationId = "netcdf_data_doGetHazardSlash")
  @ApiResponse(
      description = "Returns static curves from the NSHM NetCDF file",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/{longitude}/{latitude}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetSlash(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable @Nullable Double longitude,
      @Schema(required = true) @PathVariable @Nullable Double latitude) {
    return doGet(request, longitude, latitude, null);
  }

  static class Query {
    final Double longitude;
    final Double latitude;
    final NehrpSiteClass siteClass;

    Query(Double longitude, Double latitude, NehrpSiteClass siteClass) {
      this.longitude = longitude;
      this.latitude = latitude;
      this.siteClass = siteClass;
    }
  }

  static enum Service {
    BOUNDING,
    CURVES,
    CURVES_BY_SITE_CLASS,
  }
}
