package gov.usgs.earthquake.nshmp.netcdf.www;

import static gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.GSON;

import java.util.List;
import java.util.logging.Logger;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Bounds;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.Netcdf;
import gov.usgs.earthquake.nshmp.netcdf.data.ScienceBaseMetadata;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfController.Query;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfController.Service;
import gov.usgs.earthquake.nshmp.www.Response;

import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;

/**
 * Abstract service handler for {@code NetcdfController}.
 *
 * @see NetcdfController
 *
 * @author U.S. Geological Survey
 */
public abstract class NetcdfService {

  protected static final Logger LOGGER = Logger.getLogger(NetcdfService.class.getName());

  Netcdf<?> netcdf;

  protected NetcdfService(Netcdf<?> netcdf) {
    this.netcdf = netcdf;
  }

  /**
   * Returns the metadata response.
   *
   * @param httpRequest The HTTP request
   */
  abstract Response<String, Metadata> getMetadataResponse(HttpRequest<?> httpRequest);

  /**
   * Returns the service name
   */
  abstract String getServiceName();

  /**
   * Returns the netcdf object associated with the specific data type.
   */
  abstract Netcdf<?> netcdf();

  /**
   * Returns the static curves at a specific location.
   *
   * @param <T> The response type
   * @param request The request data
   * @param url The URL for the service call
   */
  abstract <T> Response<RequestData, T> processCurves(RequestData request, String url);

  /**
   * Returns the static curves at a specific location and site class.
   *
   * @param <T> The response type
   * @param request The request data
   * @param url The URL for the service call
   */
  abstract <T> Response<RequestDataSiteClass, T> processCurvesSiteClass(
      RequestDataSiteClass request,
      String url);

  /**
   * Process the service request and returns the reponse.
   *
   * @param httpRequest The HTTP request
   * @param query The HTTP query
   * @param service The NetCDF service
   */
  abstract Response<?, ?> processRequest(HttpRequest<?> httpRequest, Query query, Service service);

  /**
   * Handler of HTTP request, returns the service response.
   *
   * @param httpRequest The HTTP request
   * @param query The service query
   */
  HttpResponse<String> handleServiceCall(HttpRequest<?> httpRequest, Query query) {
    try {
      var url = httpRequest.getUri().toString();
      LOGGER.info("Request: " + url);
      LOGGER.fine("Query:\n" + GSON.toJson(query));

      if (query.longitude == null && query.latitude == null) {
        return metadata(httpRequest);
      }
      var service = getService(query);
      Response<?, ?> svcResponse = processRequest(httpRequest, query, service);
      var response = GSON.toJson(svcResponse);
      LOGGER.fine("Result:\n" + response);
      return HttpResponse.ok(response);
    } catch (Exception e) {
      var url = httpRequest.getUri().toString();
      return NetcdfWsUtils.handleError(e, getServiceName(), url);
    }
  }

  private Service getService(Query query) {
    if (query.siteClass != null) {
      return Service.CURVES_BY_SITE_CLASS;
    } else {
      return Service.CURVES;
    }
  }

  private HttpResponse<String> metadata(HttpRequest<?> httpRequest) {
    var svcResponse = getMetadataResponse(httpRequest);
    var response = GSON.toJson(svcResponse);
    LOGGER.fine("Result:\n" + response);
    return HttpResponse.ok(response);
  }

  class Metadata {
    final String description;
    final String[] syntax;
    final String netcdfFile;
    final Parameters parameters;
    final ScienceBaseMetadata scienceBaseMetadata;

    Metadata(HttpRequest<?> request, String description) {
      var url = request.getUri().toString();
      url = url.endsWith("/") ? url.substring(0, url.length() - 1) : url;
      this.description = description;
      scienceBaseMetadata = netcdf().netcdfData().scienceBaseMetadata();
      netcdfFile = netcdf().netcdfPath().toString();
      syntax = new String[] {
          url + "/{longitude:number}/{latitude:number}",
          url + "?longitude={number}&latitude={number}",
          url + "/{longitude:number}/{latitude:number}/{siteClass:NehrpSiteClass}",
          url + "?longitude={number}&latitude={number}&siteClass={NehrpSiteClass}",
      };
      parameters = new Parameters();
    }
  }

  class Parameters {
    final Bounds bounds;
    final List<NehrpSiteClass> siteClasses;

    Parameters() {
      var coordinates = netcdf().netcdfData();
      bounds = coordinates.bounds();
      siteClasses = coordinates.siteClasses();
    }
  }

  static class ResponseMetadata extends RequestDataSiteClass {
    final String xLabel;
    final String yLabel;

    ResponseMetadata(
        String xLabel,
        String yLabel,
        Location site,
        NehrpSiteClass siteClass) {
      super(site, siteClass);
      this.xLabel = xLabel;
      this.yLabel = yLabel;
    }
  }

  static class ResponseData<T extends ResponseMetadata> {
    final T metadata;
    final XySequence data;

    ResponseData(T metadata, XySequence data) {
      this.metadata = metadata;
      this.data = data;
    }
  }

  static class RequestData {
    Location site;

    RequestData(Location site) {
      this.site = site;
    }
  }

  static class RequestDataSiteClass extends RequestData {
    NehrpSiteClass siteClass;

    RequestDataSiteClass(Location site, NehrpSiteClass siteClass) {
      super(site);
      this.siteClass = siteClass;
    }
  }
}
