package gov.usgs.earthquake.nshmp.netcdf.www;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfGroundMotions;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;
import gov.usgs.earthquake.nshmp.netcdf.reader.BoundingReaderGroundMotions;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfController.Query;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfController.Service;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.Key;
import gov.usgs.earthquake.nshmp.www.Response;
import gov.usgs.earthquake.nshmp.www.WsUtils;
import gov.usgs.earthquake.nshmp.www.meta.Status;

import io.micronaut.http.HttpRequest;

/**
 * Handles service calls for ground motions.
 *
 * @see NetcdfController
 *
 * @author U.S. Geological Survey
 */
public class NetcdfServiceGroundMotions extends NetcdfService {

  static final String SERVICE_DESCRIPTION = "Get static ground motions from a NetCDF file";
  static final String SERVICE_NAME = "Static Ground Motions";
  static final String X_LABEL = "Period (s)";
  static final String Y_LABEL = "Median Ground Motion (g)";

  protected NetcdfServiceGroundMotions(NetcdfGroundMotions netcdf) {
    super(netcdf);
  }

  @Override
  Response<String, Metadata> getMetadataResponse(HttpRequest<?> request) {
    var metadata = new Metadata(request, SERVICE_DESCRIPTION);
    var url = request.getUri().toString();
    return new Response<>(Status.USAGE, SERVICE_NAME, url, metadata, url);
  }

  @Override
  String getServiceName() {
    return SERVICE_NAME;
  }

  @Override
  NetcdfGroundMotions netcdf() {
    return (NetcdfGroundMotions) netcdf;
  }

  @Override
  Response<?, ?> processRequest(HttpRequest<?> httpRequest, Query query, Service service) {
    var site = Location.create(query.longitude, query.latitude);
    var requestData = new RequestData(site);
    var url = httpRequest.getUri().toString();

    switch (service) {
      case CURVES:
        return processCurves(requestData, url);
      case CURVES_BY_SITE_CLASS:
        requestData = new RequestDataSiteClass(site, query.siteClass);
        return processCurvesSiteClass(
            (RequestDataSiteClass) requestData, url);
      default:
        throw new RuntimeException("Netcdf service [" + service + "] not found");
    }
  }

  @Override
  Response<RequestDataSiteClass, ResponseData<ResponseMetadata>> processCurvesSiteClass(
      RequestDataSiteClass request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.site.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.site.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    var curves = netcdf().staticData(request.site, request.siteClass);
    var responseData = toResponseData(request, curves);
    return new Response<>(Status.SUCCESS, SERVICE_NAME, request, responseData, url);
  }

  @Override
  Response<RequestData, List<ResponseData<ResponseMetadata>>> processCurves(RequestData request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.site.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.site.longitude);
    var curves = netcdf().staticData(request.site);
    var responseData = toList(request.site, curves);
    return new Response<>(Status.SUCCESS, SERVICE_NAME, request, responseData, url);
  }

  List<ResponseData<ResponseMetadata>> toList(
      Location site,
      StaticData<XySequence> curves) {
    return curves.entrySet().stream()
        .map(entry -> {
          var request = new RequestDataSiteClass(site, entry.getKey());
          return toResponseData(request, entry.getValue());
        })
        .collect(Collectors.toList());
  }

  ResponseData<ResponseMetadata> toResponseData(
      RequestDataSiteClass request,
      XySequence curves) {
    var metadata = new ResponseMetadataGroundMotions(
        X_LABEL,
        Y_LABEL,
        request.site,
        request.siteClass);
    return new ResponseData<>(metadata, curves);
  }

  static class ResponseMetadataGroundMotions extends ResponseMetadata {
    Map<String, Double> imtValues;

    ResponseMetadataGroundMotions(
        String xLabel,
        String yLabel,
        Location site,
        NehrpSiteClass siteClass) {
      super(xLabel, yLabel, site, siteClass);
      imtValues = Map.of(
          Imt.PGA.name(),
          BoundingReaderGroundMotions.PGA_VALUE,
          Imt.PGV.name(),
          BoundingReaderGroundMotions.PGV_VALUE);
    }
  }
}
