package gov.usgs.earthquake.nshmp.netcdf.www;

import java.util.List;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticDataHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfController.Query;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfController.Service;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.Key;
import gov.usgs.earthquake.nshmp.www.Response;
import gov.usgs.earthquake.nshmp.www.WsUtils;
import gov.usgs.earthquake.nshmp.www.meta.Status;

import io.micronaut.http.HttpRequest;

/**
 * Handles service calls for hazard curves.
 *
 * @see NetcdfController
 *
 * @author U.S. Geological Survey
 */
public class NetcdfServiceHazardCurves extends NetcdfService {

  static final String SERVICE_DESCRIPTION = "Get static hazard curves from a NetCDF file";
  static final String SERVICE_NAME = "Static Hazard Curves";
  static final String X_LABEL = "Ground Motion (g)";
  static final String Y_LABEL = "Annual Frequency of Exceedence";

  NetcdfServiceHazardCurves(NetcdfHazardCurves netcdf) {
    super(netcdf);
  }

  @Override
  Response<String, Metadata> getMetadataResponse(HttpRequest<?> request) {
    var metadata = new Metadata(request, SERVICE_DESCRIPTION);
    var url = request.getUri().toString();
    return new Response<>(Status.USAGE, SERVICE_NAME, url, metadata, url);
  }

  @Override
  String getServiceName() {
    return SERVICE_NAME;
  }

  @Override
  NetcdfHazardCurves netcdf() {
    return (NetcdfHazardCurves) netcdf;
  }

  @Override
  Response<RequestData, List<List<ResponseData<ResponseMetadataHazard>>>> processCurves(
      RequestData request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.site.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.site.longitude);
    var curves = netcdf().staticData(request.site);
    var curvesAsList = toList(request.site, curves);
    return new Response<>(Status.SUCCESS, SERVICE_NAME, request, curvesAsList, url);
  }

  @Override
  Response<RequestDataSiteClass, List<ResponseData<ResponseMetadataHazard>>> processCurvesSiteClass(
      RequestDataSiteClass request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.site.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.site.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    var curves = netcdf().staticData(request.site, request.siteClass);
    var curvesAsList = toList(request, curves);
    return new Response<>(Status.SUCCESS, SERVICE_NAME, request, curvesAsList, url);
  }

  @Override
  Response<?, ?> processRequest(HttpRequest<?> httpRequest, Query query, Service service) {
    var site = Location.create(query.longitude, query.latitude);
    var requestData = new RequestData(site);
    var url = httpRequest.getUri().toString();

    switch (service) {
      case CURVES:
        return processCurves(requestData, url);
      case CURVES_BY_SITE_CLASS:
        requestData = new RequestDataSiteClass(site, query.siteClass);
        return processCurvesSiteClass(
            (RequestDataSiteClass) requestData, url);
      default:
        throw new RuntimeException("Netcdf service [" + service + "] not found");
    }
  }

  List<ResponseData<ResponseMetadataHazard>> toList(
      RequestDataSiteClass request,
      StaticDataHazardCurves curves) {
    return curves.entrySet().stream()
        .map((entry) -> {
          return toResponseData(request, entry.getKey(), entry.getValue());
        })
        .collect(Collectors.toList());
  }

  List<List<ResponseData<ResponseMetadataHazard>>> toList(
      Location site,
      StaticData<StaticDataHazardCurves> curves) {
    return curves.entrySet().stream()
        .map(entry -> {
          var request = new RequestDataSiteClass(site, entry.getKey());
          return toList(request, entry.getValue());
        })
        .collect(Collectors.toList());
  }

  ResponseData<ResponseMetadataHazard> toResponseData(
      RequestDataSiteClass request,
      Imt imt,
      XySequence curves) {
    var metadata =
        new ResponseMetadataHazard(X_LABEL, Y_LABEL, request.site, request.siteClass, imt);
    return new ResponseData<>(metadata, curves);
  }

  static class ResponseMetadataHazard extends ResponseMetadata {
    final Imt imt;

    ResponseMetadataHazard(
        String xLabel,
        String yLabel,
        Location site,
        NehrpSiteClass siteClass,
        Imt imt) {
      super(xLabel, yLabel, site, siteClass);
      this.imt = imt;
    }
  }
}
